/***********************************************************************
 * PDA600 for Amstrad Handheld-Computer
 * V0.9 06.06.96 Thomas Hommes	(hm@bg.bib.de)
 * V1.0 04.09.96 add -i mode
 * V1.1 13.09.96 add -e mode
 * V1.2 22.10.96 added full MS-Office Scheduler+ compatibility for -i
 *               by: Frans Breukelman <fb@knoware.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 ***********************************************************************/
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include <ctype.h>
#include "mcomm\comm.h"

#define MEMSIZE 110720l		/* Memory Size of PDA600 */

ASYNC 	port;

/***********************************************************************
 * term()
 ***********************************************************************/
void term(void) {
  int key;
  int c,in=0;

  printf("Terminal-Program, ESC for exit\n\n\n");
  async_txflush(&port); async_rxflush(&port);
  printf("****OUT****\n");
  do
  {
	if ( kbhit() ) {
		key = getch();
		if (in) { printf("\n****OUT****\n"); in=!in; }
		if (key==28) key=0;
		if ((isgraph(key) || isspace(key)) && key!=0x15) putch(key);
			else printf("[0x%02x]",key);
		async_tx(&port,key);
	}
	if (!((c = async_rx(&port)) & B_RXEMPTY)) {
		c= c & 0xff;
		if (! in) { printf("\n****IN*****\n"); in=!in; }
		if ((isgraph(c) || isspace(c)) && c!=0x15) putch(c);
			else printf("[0x%02x]",c);
	}
   }
   while ( key != 27 );
}

/***********************************************************************
 * savedata()
 ***********************************************************************/
void savedata(char *fn) {
	int c;
	long int i,t;
	FILE *fdata;

	if ((fdata=fopen((fn!=NULL) ? fn : "PDA600.TXT","wb"))==NULL) {
		perror((fn!=NULL) ? fn : "PDA600.TXT");
		return;
	}


	printf("please start print program on PDA600... (any key to abort)\n");

	async_txflush(&port); async_rxflush(&port);
	while (! kbhit() && async_rxcnt(&port)==0);

	printf("save data to %s\n\n",(fn!=NULL) ? fn : "PDA600.TXT");
	printf("0....10...20...30...40...50...60...70...80...90...100... kByte\n");
	i=0; t=0;
	while ( ! kbhit() )
	{
		if (!((c = async_rx(&port)) & B_RXEMPTY)) {
			c= c & 0xff;
			fputc(c,fdata);
			t=0;
			if ((i++ %   2000l)==0) putch('.');
		}
		if (t++ >1000) break;
	}
	if (kbhit()) {
		printf("\nsave data aborted...!\n");
	}
	fclose(fdata);
	printf("\n\n%lu Bytes saved!\n",i);
}

/***********************************************************************
 * pda2ascii() PDA-Chars to PC8
 ***********************************************************************/
void pda2ascii(unsigned char *buf,int len)
{
    int i;
    for (i=0;i<len;i++) {
	switch (buf[i]) {
	    /*  */
	    case 0x7c: buf[i]='';  break;
	    /*  */
	    case 0x7f: buf[i]='';  break;
	    /*  */
	    case 0x5c: buf[i]='';  break;
	    /*  */
	    case 0x83: buf[i]='';  break;
	    /*  */
	    case 0x93: buf[i]='';  break;
	    /*  */
	    case 0x98: buf[i]='';  break;
	    /*  */
	    case 0x99: buf[i]='';  break;
	}
    }
}
/***********************************************************************
 * ascii2pda() PC8 to PDA and ANSI to PDA
 ***********************************************************************/
void ascii2pda(unsigned char *buf,int len)
{
    int i;
    for (i=0;i<len;i++) {
	switch (buf[i]) {
	    /*  */
	    case 153 : buf[i]=0x7c; break;
	    case 214 : buf[i]=0x7c; break;
	    /*  */
	    case 154 : buf[i]=0x7f; break;
	    case 220 : buf[i]=0x7f; break;
	    /*  */
	    case 142 : buf[i]=0x5c; break;
	    case 196 : buf[i]=0x5c; break;
	    /*  */
	    case 132 : buf[i]=0x83; break;
	    case 228 : buf[i]=0x83; break;
	    /*  */
	    case 148 : buf[i]=0x93; break;
	    case 246 : buf[i]=0x93; break;
	    /*  */
	    case 129 : buf[i]=0x98; break;
	    case 252 : buf[i]=0x98; break;
	    /*  */
	    case 225 : buf[i]=0x99; break;
	    case 223 : buf[i]=0x99; break;
	}
    }
}
/***********************************************************************
 * sendrecord()
 ***********************************************************************/
void sendrecord(unsigned char *buf, int len) {
	unsigned int i,crc;
	i=0; crc=0;

	async_tx(&port, 0x01); 		/* start 0x01 */
	async_tx(&port, len & 0xff);     /* lowbyte length */
	async_tx(&port, len >> 8);	/* highbyte length */

	while ( ! kbhit() )
	{
		if (i+1>len) break;
		async_tx(&port,buf[i]);
/*/		printf("%02X ",buf[i]);*/
		crc+=buf[i];
		i++;
	}
	async_tx(&port,crc & 0xff); 	/* send crc-byte */
	while (async_txcnt(&port)>0);
/*/	printf("type: %02X  length %04X   crc: %02X\n",buf[0],len,crc &0xff);*/
	putc('.',stdout);
}

/***********************************************************************
 * importdata()
 ***********************************************************************/
void importdata(char *fn,int type) {
    int len,i,t;
    char buf[1024],line[1024],*pline,*pbuf;
    char year[1024];
    FILE *fdata;

    if ((fdata=fopen((fn!=NULL) ? fn : "PDA600.TXT","r"))==NULL) {
	perror((fn!=NULL) ? fn : "PDA600.TXT");
	return;
    }

    printf("import data from %s\n\n",(fn!=NULL) ? fn : "PDA600.TXT");

    printf("please start import program on PDA600... (any key to abort)\n");

    /* flush buffers */
    async_txflush(&port); async_rxflush(&port);

    while (1) {
	if (fgets(line,sizeof(line),fdata)==NULL) break;
	if (strlen(line)<3) break;
	ascii2pda(line,strlen(line));

    pline=strtok(line,"\n\t;-.:,");
	memset(buf,0,sizeof(buf));
	switch (type) {
	    case 1: /* scheduler/yearly days */
	    case 3: /* action list */
		buf[0]=type;
        buf[33]=atoi(pline); pline=strtok(0,"\n\t;-.:,"); /* day */
        buf[34]=atoi(pline); pline=strtok(0,"\n\t;-.:,"); /* month */
/* - token added for MS-Office compatibility
* century check added for MS-Office compatibility
*/
        if(atoi(pline) > 1900)
           buf[35]=atoi(pline)-1900;
        else
           buf[35]=atoi(pline);
        pline=strtok(0,"\n\t;-.:,"); /* year */
		t      =atoi(pline)*100; pline=strtok(0,"\n\t;.:,"); /* hour */
		t     +=atoi(pline); pline=strtok(0,"\n\t;.:,"); /* min */
		buf[36]=t & 0xff; buf[37]=t >>8;
		strncpy(&buf[10],pline,21); pline=strtok(0,"\n\t;.:,"); /* infoline */
		buf[38]=0x05; buf[40]=0x21; buf[41]=0x01; /* const data */
		if (type==3) buf[42]=atoi(pline); /* action-flag */
		if (type==1 && atoi(pline)!=0) buf[0]=4; /* yearly day */
		len=44;
		break;
	    case 5: /* address book */
		buf[0]=0x05; buf[16]=0x10;
		pbuf=&buf[18];
		for(i=0;i<26;i++) {
		    switch (i) {
		    /* type of call number (0x0-0x5) */
			case 0:
			case 1:
			case 2:
			case 3:
				buf[10+i]= atoi(pline) > 5 ? 0 : atoi(pline);
				break;
			/* strings, n1-2,a1-5,cn1-4,ct1-4,i1-7 */
			default: strncpy(pbuf,pline,21);
				 pbuf+=strlen(pline)< 21 ? strlen(pline)+1 : 22;
		    }
		    pline=strtok(0,"\n\t;.:,");
		}
		len=pbuf-buf;
		buf[14]=(len-2) & 0xff; buf[15]=(len-2) >> 8;
		break;
	    default:
		printf("invalid record type for -l");
		exit(1);
	}
	/* wait for ACK */
	while (! kbhit() && (async_rx(&port) & 0xff) !=0x06);
	sendrecord(buf,len);
    }

    /* send NIL record at end */
    if (! kbhit()) {
	while (! kbhit() && (async_rx(&port) & 0xff) !=0x06);
	sendrecord(NULL,0);
    }

    if (kbhit()) printf("\nload aborted...!\n");
    fclose(fdata);
}
/***********************************************************************
 * readrecord()
 ***********************************************************************/
unsigned long readrecord(FILE *fdata, unsigned long pos, char *buf)
{
	unsigned char mem[5];
	unsigned long nextpos,len;
	nextpos=pos;
	do {
		pos=nextpos;
		if (kbhit()) return 0;
		if (fseek(fdata,pos,SEEK_SET)!=0) return 0;
		if (fread(mem,5,1,fdata)!=1) return 0;
		nextpos= 65536l*(mem[3] & 0x1) + 256l*mem[2] + mem[1];
		len=nextpos-pos;
/*		printf("readrecord:%02X %08lX-%08lX len: %lu\n",mem[0],pos,nextpos,len);*/
	} while (mem[0]==0x00);
	if (mem[0]!=0x01) return 0;
	len=(len<100) ? len : 2;
	fread(buf,len,1,fdata);
	pda2ascii(buf,len);
	return nextpos;
}
/***********************************************************************
 * exportdata()
 ***********************************************************************/
void exportdata(char *fn,int type) {
    int i;
    char *p;
    unsigned long pos;
    unsigned char buf[1024],*pbuf;
    FILE *fdata;

    if ((fdata=fopen((fn!=NULL) ? fn : "PDA600.BAK","rb"))==NULL) {
	perror((fn!=NULL) ? fn : "PDA600.BAK");
	return;
    }

    fprintf(stderr,"read data from %s\n\n",(fn!=NULL) ? fn : "PDA600.BAK");

    pos=0; memset(buf,0,sizeof(buf));
    while ((pos=readrecord(fdata,pos,buf))!=0) {
	if((buf[0]==type && buf[1]==0) ||
	   (type==5 && buf[3]==0 && buf[7]==0 && buf[14]==0x10 && buf[15]==0) /* address record */)
	switch (type) {
	    case 1: /* scheduler */
	    case 3: /* action list */
	    case 4: /* yearly days */
		printf("%02u.%02u.%02u;%02u:%02u;%s;%s\n",
			buf[33],buf[34],buf[35],
			(buf[36]+buf[37]*256)/100,(buf[36]+buf[37]*256) % 100,
			&buf[10],(buf[0]==4 || buf[42]==1) ? "1" : "" );
		break;
	    case 5: /* address book */
		printf("%u;%u;%u;%u",buf[8],buf[9],buf[10],buf[11]);
		p=&buf[16];
		for (i=0;i<22;i++) {
			printf(";%s",p); p+=strlen(p)+1;
		}
		printf("\n");
		break;
	    default:
/*		printf("record type %02X\n",buf[0]);
		for (i=0;i<20;i++) putc(isascii(buf[i]) ? buf[i] : '.',stdout);
		printf("\n");
*/		break;
	}
	if (kbhit()) break;
	memset(buf,0,sizeof(buf));
    }

    if (kbhit()) fprintf(stderr,"\nexport aborted...!\n");
    fclose(fdata);
}

/***********************************************************************
 * backup()
 ***********************************************************************/
void backup(char *fn) {
	int c;
	long int i,t;
	FILE *fbackup;

	if ((fbackup=fopen((fn!=NULL) ? fn : "PDA600.BAK","wb"))==NULL) {
		perror((fn!=NULL) ? fn : "PDA600.BAK");
		return;
	}


	printf("please start backup program on PDA600... (any key to abort)\n");

	async_txflush(&port); async_rxflush(&port);
	while ( ! kbhit() )				/*Ist eine Taste gedrckt*/
    {
	 if (!((c = async_rx(&port)) & B_RXEMPTY)) {
			c=c & 0xff;
/*			if (isgraph(c) || isspace(c)) putch(c);
			else printf(" [0x%02x] ",c);
*/			if (c==0x05) {
				async_tx(&port,0x06);
				break;
			}
		}
	}
	printf("save data to %s\n\n",(fn!=NULL) ? fn : "PDA600.BAK");
	printf("0....10...20...30...40...50...60...70...80...90...100... kByte\n");
	i=0; t=0;
	while ( ! kbhit() )
	{
		if (!((c = async_rx(&port)) & B_RXEMPTY)) {
			c= c & 0xff;
			fputc(c,fbackup);
			t=0;
			if ((i++ %   2000l)==0) putch('.');
		}
		if (t++ >1000) break;
	}
	if (kbhit()) {
		printf("\nbackup aborted...!\n");
	}
	fclose(fbackup);
	printf("\n\n%lu/%lu Bytes (%3u%%) saved!\n",i,MEMSIZE,i*100/MEMSIZE);
	if(i!=MEMSIZE) printf("%cbackup file has invalid size!\n",7);
}

/***********************************************************************
 * bdummy()
 ***********************************************************************/
void bdummy(char *fn) {
	int c;
	long int i;
	FILE *fbackup;

	if ((fbackup=fopen((fn!=NULL) ? fn : "PDA600.000","wb"))==NULL) {
		perror((fn!=NULL) ? fn : "PDA600.000");
		return;
	}


	async_txflush(&port); async_rxflush(&port);
	printf("save data to %s\n\n",(fn!=NULL) ? fn : "PDA600.000");
	printf("0....10...20...30...40...50...60...70...80...90...100... kByte\n");
	i=0;
	while (i<MEMSIZE )
	{
			fputc(0x00,fbackup);
			if ((i++ %   2000l)==0) putch('.');
	}
	fclose(fbackup);
	printf("\n\n%lu/%lu Bytes (%3u%%) saved!\n",i,MEMSIZE,i*100/MEMSIZE);
}

/***********************************************************************
 * restore()
 ***********************************************************************/
void restore(char *fn) {
	int c;

	long int i;
	FILE *frestore;

	if ((frestore=fopen((fn!=NULL) ? fn : "PDA600.BAK","rb"))==NULL) {
		perror((fn!=NULL) ? fn : "PDA600.BAK");
		return;
	}
	fseek(frestore,0l,SEEK_END);
	if (MEMSIZE!=ftell(frestore)) {
		printf("%s: not a backup file (file size != memory size %lu Bytes)\n",
			(fn!=NULL) ? fn : "PDA600.BAK",MEMSIZE);
		fclose(frestore);
		return;
	}
	fseek(frestore,0l,SEEK_SET);



	printf("please start restore program on PDA600... (any key to abort)\n");

	async_txflush(&port); async_rxflush(&port);
	while (! kbhit()) {
		async_tx(&port,0x05);
		sleep(1);
		if ((async_rx(&port) & 0xff) ==0x06) break;
	}

	printf("load backup from %s\n\n",(fn!=NULL) ? fn : "PDA600.BAK");
	printf("0....10...20...30...40...50...60...70...80...90...100... kByte\n");
	i=0;
	if (! kbhit()) while ((c=fgetc(frestore))!=EOF) {
			if ((i++ %   2000l)==0) putch('.');
			while (async_txcnt(&port)>0);
			async_tx(&port,c);
	} else {
		printf("\nload data aborted...!\n");
	}
	fclose(frestore);
	printf("\n\n%lu/%lu Bytes (%3u%%) restored!\n",i,MEMSIZE,i*100/MEMSIZE);
}

/***********************************************************************
 * main()
 ***********************************************************************/
int main(int argc, char *argv[])
{
	int rcode;

	if (argc<3) {
		printf("PDA600 COM1/2/3/4 [Options]    Utility for Amstrad PDA600 Handheld\n\n");
		printf("Options:  -t	               terminal-mode\n");
		printf("          -d  [file]           save data to file\n");
		printf("          -D  [file]           save data to file endless\n");
		printf("          -ix [file]           import data from file (record type x)\n");
		printf("          -ex [file]           export data from backup-file (record type x)\n");
		printf("          -r  [file]           restore memory from file\n");
		printf("          -b  [file]           backup  memory to file\n\n");
		printf("record type: 1=scheduler/yearly days, 3=action list, 5=address book\n");
		printf("format-1/3: DD.MM.YY;HH:MM;info;Flag(0/1 for normal/yearly or action do/done)\n");
		printf("format-5  : call-type(0-5)-1;-2;-3;-4;name-1;-2;address-1;-2;-3;-4;-5;\n");
		printf("            call-nr-1;-2;-3;-4;call-name-1;-2;-3;-4;info-1;-2;-3;-4;-5;-6;-7\n");
		printf("call-type : 0=not in use, 1=home, 2=office, 3=mobile, 4=fax, 5=call-name\n");
		exit(1);
	}

	clrscr();
    fprintf(stderr,"PDA600   Utility for Amstrad Handheld-Computer   V1.2 [hm@bg.bib.de] 10/96\n");
    fprintf(stderr,"                                                      [fb@knoware.nl]\n");

	/*  O P E N   T H E   C O M M   P O R T  */
	AllocRingBuffer(&port, 1000, 1000, 1);    /* alloc FAR ring bufrs */

	if 	(argv[1][strlen(argv[1])-1]=='1')
		rcode=async_open(&port, COM1 , "9600N81");
	else if (argv[1][strlen(argv[1])-1]=='2')
		rcode=async_open(&port, COM2 , "9600N81");
	else if (argv[1][strlen(argv[1])-1]=='3')
		rcode=async_open(&port, COM3 , "9600N81");
	else if (argv[1][strlen(argv[1])-1]=='4')
		rcode=async_open(&port, COM4 , "9600N81");
	else {
		fprintf(stderr,"\ninvalid COM-Port selected!\n");
		exit(1);
	}
	if (rcode!=R_OK) {
		fprintf(stderr,"\nAsync_open failed, exit code = %d", rcode);
		exit(1);
	}
	fprintf(stderr,"Port: %s selected (%s)\n\n",strupr(argv[1]),(async_16550(&port)) ? "FIFO-16550" : "Standard-8250");
	async_msrflow(&port, B_RTS);

	if (strcmp(argv[2],"-t")==0) 	term();
	if (strcmp(argv[2],"-d")==0) 	savedata(argv[3]);
	if (strcmp(argv[2],"-D")==0) 	while (! kbhit()) savedata(argv[3]);
	if (strncmp(argv[2],"-i",2)==0) importdata(argv[3],atoi(&argv[2][2]));
	if (strncmp(argv[2],"-e",2)==0) exportdata(argv[3],atoi(&argv[2][2]));
	if (strcmp(argv[2],"-r")==0) 	restore(argv[3]);
	if (strcmp(argv[2],"-b")==0) 	backup(argv[3]);
	if (strcmp(argv[2],"-B")==0) 	bdummy(argv[3]);

	async_close(&port);                 /* close the port */
	return 0;
}
/***********************************************************************
 * end
 ***********************************************************************/
