#include <sys/types.h>
#include <sys/stat.h>
#include <linux/serial.h>
#include <sys/time.h>
#include <termios.h>
#include <assert.h>
#include <fcntl.h>
#include <stdio.h>

void press_return() {
  printf("Press return");
  scanf("\n");
  printf("Thanks\n");
}

void
PrintByte(char c) {
  int q;
  c &= 0xff;
  q=c;
  if ( isgraph(c) || isspace(c)) printf("%c",c);
    else printf("0x%2x ",q);
}

int pda;

void
WaitForDTR() {
  int ModemInfo; 
  printf("Waiting for DTR..."); fflush(stdout);
  do {
    ioctl(pda,TIOCMGET,&ModemInfo);
    /* printf("ModemInfo = hex %X\n",ModemInfo); */
  } while ( ! (ModemInfo & TIOCM_CAR));
  printf("Done\n");
}

int
ReadWithTimeout(int seconds, int fd, char* buffer, int size) {
  int numfds = 15;
  int fds;
  fd_set readfds;
  fd_set writefds;
  fd_set exceptfds; 
  struct timeval timeout;
  int BytesRead=0;

  FD_ZERO(&writefds);
  FD_ZERO(&exceptfds);

  timeout.tv_sec = seconds;
  timeout.tv_usec = 0;
    
  FD_ZERO(&readfds);
  FD_SET(fd,&readfds);

  fds = select(numfds,&readfds,&writefds,&exceptfds,&timeout);
  assert(fds != -1);

  if (FD_ISSET(fd,&readfds)) {
    BytesRead = read(fd,buffer,size);
  }
  return BytesRead;
}

#define BUFFERSIZE 1000
void DumpToFile(char *DumpFileName) {
  char buffer[BUFFERSIZE];
  int BytesRead;

  int DumpFile = creat(DumpFileName,0666);	assert(DumpFile != -1);
  do {
    BytesRead = ReadWithTimeout(3, pda, buffer, BUFFERSIZE),
    write(DumpFile,buffer,BytesRead);
  } while (BytesRead > 0);
  close(DumpFile);
}

void
backup(char * DumpFileName) {
  char c;
  WaitForDTR();
  /* printf("Waiting for 5\n"); */
  do {
    read(pda,&c,1);
    /* printf("Got %d\n",c); */
  } while (c != 5);
  write(pda,"\006",1);
  printf("Dumping...\n");
  DumpToFile(DumpFileName);
}

void
restore(char *DumpFileName) {
  char c = '\000';
  int BytesRead;
  FILE *DumpFile = fopen(DumpFileName,"r");	assert(DumpFile != NULL);
  WaitForDTR();
  /* printf("Waiting for 6\n"); */
  do {
    write(pda,"\005",1);
    /* printf("Wrote 0x05\n"); */
    BytesRead = ReadWithTimeout(3, pda, &c, 1);
    /* printf("Got %d\n",c); */
  } while (c != 6);
  printf("Restoring...\n");
  while (! feof(DumpFile)) {
    char c = fgetc(DumpFile);
    write(pda,&c,1);
  }
  close(DumpFile);
}

void
SaveData(char *DataFileName) {
  char c;
  WaitForDTR();
  printf("Saving...\n");
  DumpToFile(DataFileName);
}

void
LoadData(char *DataFileName) {
  char c;
  FILE *DataFile = fopen(DataFileName,"r");	assert(DataFile != NULL);
  WaitForDTR();
  do {
    read(pda,&c,1);
  } while (c != 6);
  write(pda,"\005",1);
  printf("Loading...\n");
  while (! feof(DataFile)) {
    char c = fgetc(DataFile);
    write(pda,&c,1);
  }
  close(DataFile);
}

void
print_usage(char *argv0) {
  printf("Usage: %s -b [backupfilename]\n",argv0);
  printf("       %s -r [restorefilename]\n",argv0);
  printf("       %s -s [savefilename]\n",argv0);
  /* printf("       %s -l [loadfilename]\n",argv0); */
  exit(1);
}

main(int argc, char * argv[]) {
  char *DeviceName = "/dev/pda600";
  char *DumpFileName = "PDA600.dump";
  char *DataFileName = "PDA600.data";
  int i; 
  struct serial_struct SerialInfo;
  char c;


  if ((argc < 2) || (argc > 3)) print_usage(argv[0]);
  if (argv[1][0] != '-') print_usage(argv[0]);

  pda = open(DeviceName,O_RDWR); 	assert(pda != -1);

  { /* Set Baudrate etc */
    int res;
    struct termios tio;
    res = tcgetattr(pda,&tio);		assert(res == 0);

    res = cfsetospeed(&tio,B9600);	assert(res == 0);
    res = cfsetispeed(&tio,B9600);	assert(res == 0);

    tio.c_iflag |= BRKINT;
    tio.c_iflag &= ~ICRNL;
    tio.c_iflag &= ~IMAXBEL;
    tio.c_iflag &= ~IXOFF;
    tio.c_iflag &= ~IXON;

    tio.c_oflag &= ~OPOST;

    tio.c_lflag &= ~ISIG;
    tio.c_lflag &= ~ICANON;
    tio.c_lflag &= ~ECHO;
    tio.c_lflag &= ~ECHOE;
    tio.c_lflag &= ~ECHOK;
    tio.c_lflag |= NOFLSH;
    tio.c_lflag |= XCASE;
    tio.c_lflag |= TOSTOP;
    tio.c_lflag |= ECHOPRT;

    res = tcsetattr(pda,TCSANOW,&tio);	assert(res == 0);
  }

  ioctl(pda,TIOCGSERIAL,&SerialInfo);
  SerialInfo.flags |= TIOCM_RTS;
  SerialInfo.flags |= TIOCM_DTR;
  SerialInfo.flags |= TIOCM_DSR;
  ioctl(pda,TIOCSSERIAL,&SerialInfo);
  
  switch (argv[1][1]) {
    case 'b': 
      if (argc  == 3) DumpFileName = argv[2];
      backup(DumpFileName);
      break;
    case 'r': 
      if (argc  == 3) DumpFileName = argv[2];
      restore(DumpFileName);
      break;
    case 's': 
      if (argc  == 3) DataFileName = argv[2];
      SaveData(DataFileName);
      break;
    case 'l': 
      if (argc  == 3) DataFileName = argv[2];
      LoadData(DataFileName);
      break;
    default: print_usage(argv[0]);
  }
  close(pda);
}
